from searoute.classes.area_feature import AreaFeature
from searoute.classes.ports_props import PortProps
from searoute.tests.test_utils import get_eur_like_poly, get_suisse_poly

import searoute as sr
import unittest

class TestPortProps(unittest.TestCase):
    def test_initialization(self):
        """Test that a PortProps object is initialized with the correct values."""
        pp = PortProps('port1', 2, {'prop1': 'value1'})
        self.assertEqual(pp.port_id, 'port1')
        self.assertEqual(pp.share, 2)
        self.assertEqual(pp.props, {'prop1': 'value1'})

    def test_default_values(self):
        """Test that default values are correctly applied."""
        pp = PortProps('port2')
        self.assertEqual(pp.port_id, 'port2')
        self.assertEqual(pp.share, 1)
        self.assertIsNone(pp.props)

    def test_properties(self):
        """Test that property getters work as expected."""
        pp = PortProps('port3', 3, {'prop2': 'value2'})
        self.assertEqual(pp.port_id, 'port3')
        self.assertEqual(pp.share, 3)
        self.assertEqual(pp.props, {'prop2': 'value2'})

    def test_immutability(self):
        """Test that the PortProps object is immutable."""
        pp = PortProps('port4', 4, {'prop3': 'value3'})
        with self.assertRaises(AttributeError):
            pp.port_id = 'new_port'
        with self.assertRaises(AttributeError):
            pp.share = 5
        with self.assertRaises(AttributeError):
            pp.props = {'prop4': 'value4'}


    def test_in_list(self):
        port_props_list = [
            PortProps('port1', 2, {'prop1': 'value1'}),
            PortProps('port2', 3, {'prop2': 'value2'}),
            PortProps('port3', 4, {'prop3': 'value3'}),
            # Add more PortProps instances as needed
        ]
        third_port_props = port_props_list[2]
        expected_share_value = 4
        self.assertEqual(third_port_props.share, expected_share_value)
        self.assertNotEqual(third_port_props.share, 1)
        self.assertEqual(port_props_list[1].port_id, 'port2')


    def test_preferred_ports(self):
        # assuming 2 ports prefered
        first_port = PortProps('FRMAR', 2)
        second_port = PortProps('NEW_PRT_ID', 7, {'x':1, 'y':2})

        # point is in Europe
        point = (7, 47)
        # europan area
        eurArea = AreaFeature(get_eur_like_poly(), 
                              name= 'EUR', 
                              preferred_ports=
                              [first_port, second_port])
        

        result = sr.setup_P().get_preferred_ports(*point, AreaFeature.create([eurArea]), top=None, include_area_name = True)
        
        self.assertEqual(2, len(result))
        # get_preferred_ports should return 1st the highest share of preferred ports, here is NEW_PRT_ID with share 7
        self.assertEqual(second_port.port_id, result[0][0])
        self.assertEqual(eurArea.properties['name'], result[0][3])

        # assuming 2 areas
        # it should select the smalles base on its area, here 'CH'
        chArea = AreaFeature(get_suisse_poly(), 
                              name= 'CH', 
                              preferred_ports=
                              [first_port])

        result = sr.setup_P().get_preferred_ports(*point, AreaFeature.create([eurArea, chArea]), top=None, include_area_name = True)
        
        self.assertEqual(chArea.properties['name'], result[0][3])

        # if point elsevere, like Paris
        point = (2.333333, 48.866667)
        result = sr.setup_P().get_preferred_ports(*point, AreaFeature.create([eurArea, chArea]), top=None, include_area_name = True)
        
        self.assertNotEqual(chArea.properties['name'], result[0][3])
        self.assertEqual(eurArea.properties['name'], result[0][3])
        
        


import geojson
from searoute.classes.area_feature import AreaFeature
from searoute.classes.ports_props import PortProps
from searoute.tests.test_utils import generate_random_word, get_eur_like_poly, get_suisse_poly


def test_norm_preferred_ports():
    area = AreaFeature(coordinates = geojson.utils.generate_random("Polygon").coordinates,
                name=generate_random_word(4), 
                preferred_ports=None)
                
    # Test case 1: None input
    result = area.norm_preferred_ports(None)
    assert result == []

    # Test case 2: Single PortProps input
    port_props_instance = PortProps("8080")
    result = area.norm_preferred_ports(port_props_instance)
    print(result)
    assert result == [port_props_instance]

    # Test case 3: String input
    result = area.norm_preferred_ports("9090")
    print(result)
    assert result == [PortProps("9090")]

    # Test case 4: Integer input
    result = area.norm_preferred_ports(7070)
    print(result)
    assert result == [PortProps("7070")]

    # Test case 5: Tuple input
    result = area.norm_preferred_ports((6060,1))
    print(result)
    assert result == [PortProps("6060", 1)]

    # Test case 6: List input with mixed types
    input_list = ["8080", 9090, (6060,), None, "7070"]
    result = area.norm_preferred_ports(input_list)
    print(result)
    assert result == [
        PortProps("8080", 1),
        PortProps("9090", 1),
        PortProps("6060",1),
        PortProps("7070", 1)
    ]
    
    # Test case 7: Unsupported type input
    result = area.norm_preferred_ports(True)
    assert result == [PortProps(True, 1)]

def test_contains():
    # Test if Basel is in Switzerland
    basel_point = (7.6174, 47.5186) # Basel city
    areaCH = AreaFeature(coordinates=get_suisse_poly(), name='CH_poly_area')
    assert True == areaCH.contains(*basel_point)

    # test if Paris is not in Switweland
    paris_point = (2.333333, 48.866667) # Paris
    assert False == areaCH.contains(*paris_point)

    # test example simplified within europe Area
    example_zone_coords = get_eur_like_poly()
    areaEUR = AreaFeature(coordinates = example_zone_coords, name='example_poly_area')
    
    # Paris in EUR
    assert True == areaEUR.contains(*paris_point)
    
    # Basel in EUR
    assert True == areaEUR.contains(*basel_point)
    
    # first 10 point of CH in EUR
    for ch_p in areaCH.geometry.coordinates[0][:10]:
         assert True == areaEUR.contains(*ch_p)

    # New york is not in EUR
    new_york_point = (-74.005941 , 40.712784 )
    assert False == areaEUR.contains(*new_york_point)

    # Tokyo is not in EUR
    tokyo_point = (139.839478, 35.652832 )
    assert False == areaEUR.contains(*tokyo_point)





from geojson import Feature, FeatureCollection, Polygon
from typing import Union, List, Tuple
from .ports_props import PortProps
from ..utils import pnpoly


class AreaFeature(Feature):

    @staticmethod
    def create(areas:list) ->List:
        """
        Create a feature collection of geojson type from a list of AreaFeature

        Parameters:
        ==========
        - areas: a list of AreaFeature

        Returns:
        ========
        FeatureCollection
        
        """
        if areas is None:
            raise Exception('Areas can not be empty')
        
        features = []
        for area in areas:
            features.append(area)

        return FeatureCollection(features)
        
        
    def __init__(self, coordinates, preferred_ports:list=None, name=None, **kwargs):
        geometry = Polygon(coordinates)
        super().__init__(geometry=geometry, properties={})

        self.properties['preferred_ports'] = self.norm_preferred_ports(preferred_ports)
        self.properties['name'] = name
        self.properties['area'] = self.calculate_geometry_area(coordinates[0])
        self.properties.update(kwargs)
        self.type = 'Feature'

    
    
    def norm_preferred_ports(self, preferred_ports: Union[None, PortProps, str, int, Tuple, List]) -> List[PortProps]:
        if preferred_ports is None:
            return []
        elif isinstance(preferred_ports, PortProps):
            return [preferred_ports]
        elif isinstance(preferred_ports, str):
            return [PortProps(preferred_ports)]
        elif isinstance(preferred_ports, int):
            return [PortProps(str(preferred_ports))]
        elif isinstance(preferred_ports, tuple):
            return [PortProps(*preferred_ports)]
        elif isinstance(preferred_ports, list):
            # Flatten the list by iterating over each item
            result = []
            for p in preferred_ports:
                result.extend(self.norm_preferred_ports(p))  # Use extend to avoid nested lists
            return result
        else:
            return [PortProps(preferred_ports)]


    def calculate_geometry_area(self, polyline_coordinates):
        """
        Calculate the area of a polygon defined by a set of coordinates using the shoelace formula.
    
        Parameters:
        - polyline_coordinates (list): List of coordinates defining the polygon.
    
        Returns:
        - float: The calculated area in square units.
        """
    
        # Ensure the polyline is closed to form a polygon
        if polyline_coordinates[0] != polyline_coordinates[-1]:
            polyline_coordinates.append(polyline_coordinates[0])
    
        # Calculate the area using the shoelace formula
        n = len(polyline_coordinates)
        area = 0.5 * sum(
            (polyline_coordinates[i][0] * polyline_coordinates[i + 1][1] -
             polyline_coordinates[i + 1][0] * polyline_coordinates[i][1])
            for i in range(n - 1)
        )
    
        # Return the absolute value of the calculated area
        return abs(area)
        

    def contains(self, x: float, y: float) -> bool:
        # will ignore other group of coordinates
        # assuming a closed and solid polygon
        poly = self.geometry.coordinates[0]
        n = len(poly)
        vx, vy  = list(zip(*poly))
        return pnpoly(n, vx, vy, x, y) 
        


import searoute as sr


def test_passages():
    traj = sr.searoute([52.99, 25.01], [-61.87, 17.15], append_orig_dest=True, restrictions=['northwest', 'chili'], return_passages=True)
    result_expected = ['suez', 'ormuz', 'babalmandab', 'gibraltar'] 
    true_result = traj['properties']['traversed_passages']
    assert sorted(result_expected)==sorted(true_result)


class Passage(str):
    babalmandab = 'babalmandab'
    bosporus = 'bosporus'
    gibraltar = 'gibraltar'
    suez = 'suez'
    panama = 'panama'
    ormuz = 'ormuz'
    northwest = 'northwest'
    malacca = 'malacca'
    sunda = 'sunda'
    chili = 'chili'
    south_africa = 'south_africa'


    @classmethod
    def valid_passages(cls):
        return {value for key, value in cls.__dict__.items() if not key.startswith('__') and not callable(value)}
    
    @classmethod
    def filter_valid_passages(cls, potential_values):
        valid_passages = cls.valid_passages()
        unique_passages = set()
        for value in potential_values:
            if value in valid_passages:
                unique_passages.add(value)
        return list(unique_passages)



import operator

class PortProps(tuple):
    port_id = property(operator.itemgetter(0))
    share = property(operator.itemgetter(1))
    props = property(operator.itemgetter(2))

    def __new__(self, port_id, share = 1, props = None):
        return tuple.__new__(PortProps, (str(port_id), share, props))


import networkx as nx

from ..utils import load_from_geojson
from .kdtree import KDTree
from . import area_feature
from geojson import FeatureCollection


class Ports(nx.Graph):
    """
    Base class for Ports network is an undirected graph. 
s
    A Ports graph stores nodes and edges with optional data, or attributes.

    Nodes are identified by an id of tuple representing a spacial location
    with (lon, lat) and it's attributes if applied.
    They form ports located around the world.


    """

    def __init__(self):
        super().__init__()
        DEFAULT_CRF = 'EPSG:3857'
        self.graph['crs'] = DEFAULT_CRF  # CRS attribute for the graph
        self.kdtree = KDTree()

    def add_node(self, node, **attr):
        if not isinstance(node, tuple):
            raise TypeError(
                "Node must be a str representing coordinates of a port.")
        x, y = node
        attr['x'] = x
        attr['y'] = y
        if not (attr['port'] and attr['cty']):
            raise TypeError(
                "Node port requires to have both port name (name), and country (cty) in properties to be correctly mapped")

        self.kdtree.add_point(node)
        super().add_node(node, **attr)


    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def subgraph(self, nodes):

        subg = super().subgraph(nodes)
        subg.kdtree = KDTree(nodes)

        return subg

    def query(self, terminals: bool = True, cty: str = None, to_cty: str = None, strict = False):
        """
        Query the Port graph

        Parameters
        ----------
                
        terminal: boolean, default True
            filters out only terminal marked ports
        cty : str, default None
            filters out ports in given country code (must be an ISO country code with 2 letters uppercase)
        to_cty: str, default None
            filters out ports that have to_country the port of discharge (country_pod)
            must be an ISO country code with 2 letters uppercase

        Returns
        -------
        A subgraph of Ports filtered
        """

        if not terminals and not cty and not to_cty:
            return self

        def cty_filter(data, cty):
            if not cty:
                return True
            else:
                return data.get('port')[:2] == cty
            
        def cty_to_filter(data, to_cty):
            if not to_cty:
                return True
            else:
                return to_cty in (data.get('to_cty') or [])

        terminal_filter = True if terminals else None

        if strict:
            strict_filter = [n for n, data in self.nodes(data=True) 
             if data.get('t') == terminal_filter and cty_filter(data, cty) and cty_to_filter(data, to_cty)]
            
            if len(strict_filter)<1:
                raise KeyError(f'There is no ports for your query terminals:{terminals}, from country:{cty}, to country:{to_cty}, strict:{strict}')
            
            return self.subgraph(strict_filter)

        # add filter terminals
        filtered_nodes = [(n, data) for n, data in self.nodes(data=True) if data.get('t') == terminal_filter]
        filtered_nodes_prev = None

        if len(filtered_nodes)>0:
            # add filter "from county"
            filtered_nodes_prev = filtered_nodes
            filtered_nodes = [(n, data) for n, data in filtered_nodes_prev if cty_filter(data, cty)]

        if len(filtered_nodes)>0:
            # add filter "to county"
            filtered_nodes_prev = filtered_nodes
            filtered_nodes = [(n, data) for n, data in filtered_nodes_prev if cty_to_filter(data, to_cty)]

        # finally bring all 
        if len(filtered_nodes)==0:
            filtered_nodes = self.nodes(data=True)
            

        if len(filtered_nodes)>0:
            return self.subgraph([n for n,_ in filtered_nodes])
        else:
            raise KeyError(f'There is no ports for your query terminals:{terminals}, from country:{cty}, to country:{to_cty}, strict:{strict}')

       

    def filter_only_terminals(self, u, v, edge_data):
        return True if edge_data.get('t', 0) == 1 else False

    def load_geojson(self, path):
        return load_from_geojson(self, path)

    @staticmethod
    def from_geojson(path):
        return Ports().load_geojson(path)
    

    def add_edges_from_list(self, edge_list):
        if not edge_list:
            return
        
        for edge in edge_list:
            u,v,args = edge
            self.add_edge(u, v, **args)

    def add_nodes_from_list(self, node_list):
        if not node_list:
            return

        for node in node_list:
            n, args = node
            self.add_node(n, **args)

    def update_kdtree(self, nodes = None):
        if nodes:
            self.kdtree = KDTree(nodes)
        else:
            self.kdtree = KDTree(self._node)

    

    def get_preferred_ports(self, x,y, ft:FeatureCollection, top = None, include_area_name = False):
        preferred_ports = []
        
        s_area = None
        s_area_size = float('inf')
        s_area_name = None
        
        _sum = 0
        for feature in ft.features:
            if not isinstance(feature, area_feature.AreaFeature):
                continue
            
            if feature.contains(x,y):
                prop = feature.properties
                
                name = prop.get('name', None)
                area = prop.get('area', float("inf"))

                if s_area_size >= area:
                    s_area_size = area
                    s_area = prop
                    s_area_name = name
                
        preferred_ports = s_area.get('preferred_ports', [])
        
        for p in preferred_ports:
            _sum+=p.share

        def _update_props(port_id, props):
            # update props
            if props is None or props == {}:
                try:
                    res = [y for x,y in self.nodes(data=True) if y['port']==port_id]
                    if len(res)>0:
                        return res[0] # first values
                except:
                    pass
            return props 

        if include_area_name:
            sorted_by_second = sorted([(a, b/max(_sum, 1), _update_props(a, c), s_area_name) for a,b,c in preferred_ports], key=lambda tup: tup[1], reverse=True)
        else:
            sorted_by_second = sorted([(a, b/max(_sum, 1), _update_props(a, c)) for a,b,c in preferred_ports], key=lambda tup: tup[1], reverse=True)
        
        if top:
            return sorted_by_second[:top]
        else:
            return sorted_by_second

    